import pytest

import numpy as np
import pandas as pd
from scipy.special import logit

from xgboost_distribution.distributions import NegativeBinomial


@pytest.fixture
def negative_binomial():
    return NegativeBinomial()


@pytest.mark.parametrize(
    "y, params, natural_gradient, expected_grad",
    [
        (
            np.array([1]),
            np.array([[np.log(2), logit(1.0)]]),
            True,
            np.array([[-1, 0.5]]),
        ),
    ],
)
def test_gradient_calculation(
    negative_binomial, y, params, natural_gradient, expected_grad
):
    grad, hess = negative_binomial.gradient_and_hessian(
        y, params, natural_gradient=natural_gradient
    )
    np.testing.assert_array_equal(grad, expected_grad)


def test_target_validation(negative_binomial):
    valid_target = np.array([0, 1, 4, 5, 10])
    negative_binomial.check_target(valid_target)


@pytest.mark.parametrize(
    "invalid_target",
    [np.array([-0.1, 1.2]), pd.Series([1.1, 0.4, 2.3])],
)
def test_target_validation_raises(negative_binomial, invalid_target):
    with pytest.raises(ValueError):
        negative_binomial.check_target(invalid_target)



import pytest

import numpy as np

from xgboost_distribution.distributions import Laplace


@pytest.fixture
def laplace():
    return Laplace()


@pytest.mark.parametrize(
    "y, params, natural_gradient, expected_grad",
    [
        (
            np.array([0, 0]),
            np.array([[0, 1], [1, 0]]),
            True,
            np.array([[0, 1], [1, 0]]),
        ),
        (
            np.array([0, 0]),
            np.array([[0, 1], [1, 0]]),
            False,
            np.array([[0, 1], [1, 0]]),
        ),
    ],
)
def test_gradient_calculation(laplace, y, params, natural_gradient, expected_grad):
    grad, hess = laplace.gradient_and_hessian(
        y, params, natural_gradient=natural_gradient
    )
    np.testing.assert_array_equal(grad, expected_grad)


def test_loss(laplace):
    loss_name, loss_value = laplace.loss(
        # fmt: off
        y=np.array([1, ]),
        params=np.array([[1, np.log(1)], ]),
    )
    assert loss_name == "LaplaceError"
    np.testing.assert_approx_equal(loss_value, -np.log(0.5))



"""Exponential distribution
"""
import numpy as np
from scipy.stats import expon

from xgboost_distribution.distributions.base import BaseDistribution
from xgboost_distribution.distributions.utils import check_is_ge_zero


class Exponential(BaseDistribution):
    """Exponential distribution with log score

    Definition:

        f(x) = 1 / scale * e^(-x / scale)

    We reparameterize scale -> log(scale) = a to ensure scale >= 0. Gradient:

        d/da -log[f(x)] = d/da -log[1/e^a e^(-x / e^a)]
                        = 1 - x e^-a
                        = 1 - x / scale

    The Fisher information = 1 / scale^2, when reparameterized:

        1 / scale^2 = I ( d/d(scale) log(scale) )^2 = I ( 1/ scale )^2

    Hence we find: I = 1

    """

    @property
    def params(self):
        return ("scale",)

    def check_target(self, y):
        check_is_ge_zero(y)

    def gradient_and_hessian(self, y, params, natural_gradient=True):
        """Gradient and diagonal hessian"""

        (scale,) = self.predict(params)

        grad = np.zeros(shape=(len(y), 1))
        grad[:, 0] = 1 - y / scale

        if natural_gradient:
            fisher_matrix = np.ones(shape=(len(y), 1, 1))

            grad = np.linalg.solve(fisher_matrix, grad)
            hess = np.ones(shape=(len(y), 1))  # we set the hessian constant
        else:
            hess = -(grad - 1)

        return grad, hess

    def loss(self, y, params):
        scale = self.predict(params)
        return "ExponentialError", -expon.logpdf(y, scale=scale).mean()

    def predict(self, params):
        log_scale = params
        scale = np.exp(log_scale)
        return self.Predictions(scale=scale)

    def starting_params(self, y):
        return (np.log(np.mean(y)),)



import pytest

import numpy as np
import pandas as pd

from xgboost_distribution.distributions.utils import (
    check_is_ge_zero,
    check_is_gt_zero,
    check_is_integer,
)


@pytest.mark.parametrize(
    "x",
    [np.array([0, 1]), pd.Series([-1, 0, 2])],
)
def test_check_is_integer(x):
    check_is_integer(x)


@pytest.mark.parametrize(
    "x",
    [np.array([0.1, 1.2]), pd.Series([-1.1, 0.4, 2.3])],
)
def test_check_is_integer_raises(x):
    with pytest.raises(ValueError):
        check_is_integer(x)


@pytest.mark.parametrize(
    "x",
    [np.array([0, 1]), pd.Series([0, 1.2, 2])],
)
def test_check_is_ge_zero(x):
    check_is_ge_zero(x)


@pytest.mark.parametrize(
    "x",
    [np.array([-0.1, 1.2]), pd.Series([-1.1, 0.4, 2.3])],
)
def test_check_is_ge_zero_raises(x):
    with pytest.raises(ValueError):
        check_is_ge_zero(x)


@pytest.mark.parametrize(
    "x",
    [np.array([1, 2]), pd.Series([0.1, 1.2, 2.2])],
)
def test_check_is_gt_zero(x):
    check_is_gt_zero(x)


@pytest.mark.parametrize(
    "x",
    [np.array([0, 1]), pd.Series([-1.1, 0.4, 2.3])],
)
def test_check_is_gt_zero_raises(x):
    with pytest.raises(ValueError):
        check_is_gt_zero(x)



"""Normal distribution
"""
import numpy as np
from scipy.stats import norm

from xgboost_distribution.distributions.base import BaseDistribution


class Normal(BaseDistribution):
    """Normal distribution with log scoring

    Definition:

        f(x) = exp( -[ (x-mean) / std ]^2 / 2 ) / std

    We reparameterize:

        a = mean         |  a = mean
        b = log ( std )  |  e^b = std

    (Note: reparameterizing to log(std) ensures that std >= 0, regardless of
    what the xgboost booster internally outputs, as std = e^b > 0.)

    The gradients are:

        d/da -log[f(x)] = e^(-2b) * (x-a) = (x-a) / var
        d/db -log[f(x)] = 1 - e^(-2b) * (x-a)^2 = 1 - (x-a)^2 / var

    as var = std^2 = e^(2b)

    The Fisher Information (diagonal):

        I(mean) = 1 / var
        I(std) = 2 / var

    In reparameterized form, we find I_r:

        1 / var = I_r [ d/d(mean) mean ]^2 = I
        2 / var = I_r [ d/d(std) log(std) ]^2 = I ( 1/(std) )^2

    Hence the reparameterized Fisher information:

        [ 1 / var, 0 ]
        [ 0,       2 ]

    Ref:

        https://www.wolframalpha.com/input/?i=d%2Fda+-log%28%28e%5E%28-%5B%28x-a%29%2Fe%5Eb%29%5D%5E2+%2F+2%29+%2F+e%5Eb%29%29
        https://www.wolframalpha.com/input/?i=d%2Fdb+-log%28%28e%5E%28-%5B%28x-a%29%2Fe%5Eb%29%5D%5E2+%2F+2%29+%2F+e%5Eb%29%29

    """

    @property
    def params(self):
        return ("loc", "scale")

    def gradient_and_hessian(self, y, params, natural_gradient=True):
        """Gradient and diagonal hessian"""

        loc, log_scale = self._split_params(params)
        var = np.exp(2 * log_scale)

        grad = np.zeros(shape=(len(y), 2))
        grad[:, 0] = (loc - y) / var
        grad[:, 1] = 1 - ((y - loc) ** 2) / var

        if natural_gradient:
            fisher_matrix = np.zeros(shape=(len(y), 2, 2))
            fisher_matrix[:, 0, 0] = 1 / var
            fisher_matrix[:, 1, 1] = 2

            grad = np.linalg.solve(fisher_matrix, grad)

            hess = np.ones(shape=(len(y), 2))  # we set the hessian constant
        else:
            hess = np.zeros(shape=(len(y), 2))  # diagonal elements only
            hess[:, 0] = 1 / var
            hess[:, 1] = 2 * ((y - loc) ** 2) / var

        return grad, hess

    def loss(self, y, params):
        loc, scale = self.predict(params)
        return "NormalError", -norm.logpdf(y, loc=loc, scale=scale).mean()

    def predict(self, params):
        loc, log_scale = self._split_params(params)
        # log_scale = np.clip(log_scale, -100, 100)  # TODO: is this needed?
        scale = np.exp(log_scale)

        return self.Predictions(loc=loc, scale=scale)

    def starting_params(self, y):
        return np.mean(y), np.log(np.std(y))

    def _split_params(self, params):
        """Return loc and log_scale from params"""
        return params[:, 0], params[:, 1]



"""Exeperiments of XGBDistribution vs NGBRegressor and XGBRegressor
"""

import glob
import logging
import os
import shutil
import sys
import tempfile
import time
import zipfile
from argparse import ArgumentParser
from dataclasses import dataclass
from datetime import datetime
from functools import partial, wraps
from pathlib import Path

import numpy as np
import pandas as pd
import requests
import sqlalchemy as sa
from ngboost import NGBRegressor
from scipy.stats import norm
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import KFold, train_test_split
from xgboost import XGBRegressor

from xgboost_distribution import XGBDistribution

_logger = logging.getLogger(__name__)


# -------------------------------------------------------------------------------------
# Datasets defintions
# -------------------------------------------------------------------------------------

DATASETS = {}  # each instantiated dataset will be stored here, keyed by name
DATA_DIR = Path(__file__).parent.parent.absolute().joinpath("data")


@dataclass
class Dataset:
    name: str
    url: str
    file_to_unpack: str = None  # if url refers to zip, we mark the file to unpack
    load_func: callable = pd.read_csv
    processing_func: callable = None

    def __post_init__(self):
        DATASETS.update({self.name: self})


Dataset(
    name="housing",
    url="https://archive.ics.uci.edu/ml/machine-learning-databases/housing/housing.data",  # noqa: E501
    load_func=partial(pd.read_csv, header=None, delim_whitespace=True),
)

Dataset(
    name="concrete",
    url="https://archive.ics.uci.edu/ml/machine-learning-databases/concrete/compressive/Concrete_Data.xls",  # noqa: E501
    load_func=pd.read_excel,
)

Dataset(
    name="energy",
    url="https://archive.ics.uci.edu/ml/machine-learning-databases/00242/ENB2012_data.xlsx",  # noqa: E501
    load_func=partial(pd.read_excel, usecols=[f"X{i}" for i in range(1, 9)] + ["Y1"]),
)

Dataset(
    name="naval",
    url="https://archive.ics.uci.edu/ml/machine-learning-databases/00316/UCI%20CBM%20Dataset.zip",  # noqa: E501
    file_to_unpack="data.txt",
    load_func=partial(pd.read_csv, header=None, delim_whitespace=True),
    processing_func=lambda x: x.iloc[:, :-1],
)

Dataset(
    name="power",
    url="https://archive.ics.uci.edu/ml/machine-learning-databases/00294/CCPP.zip",
    file_to_unpack="Folds5x2_pp.xlsx",
    load_func=pd.read_excel,
)

Dataset(
    name="protein",
    url="https://archive.ics.uci.edu/ml/machine-learning-databases/00265/CASP.csv",
    processing_func=lambda x: x.loc[
        :, ["F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8", "F9", "RMSD"]
    ],
)

Dataset(
    name="wine",
    url="https://archive.ics.uci.edu/ml/machine-learning-databases/wine-quality/winequality-red.csv",  # noqa: E501
    load_func=partial(pd.read_csv, delimiter=";"),
)

Dataset(
    name="yacht",
    url="http://archive.ics.uci.edu/ml/machine-learning-databases/00243/yacht_hydrodynamics.data",  # noqa: E501
    load_func=partial(pd.read_csv, header=None, delim_whitespace=True),
)

Dataset(
    name="msd",
    url="https://archive.ics.uci.edu/ml/machine-learning-databases/00203/YearPredictionMSD.txt.zip",  # noqa: E501
    file_to_unpack="YearPredictionMSD.txt",
    load_func=pd.read_csv,
    processing_func=lambda x: x.iloc[:, ::-1],
)

# -------------------------------------------------------------------------------------
# Datasets loading functions
# -------------------------------------------------------------------------------------


def load_dataset(name, data_dir=DATA_DIR):
    dataset = DATASETS[name]
    local_path = f"{data_dir}/{name}.data"

    if os.path.exists(local_path):
        _logger.info(f"Loading dataset from local {local_path}...")
        return load_local_dataset(dataset, local_path)
    else:
        _logger.info("Dataset not locally cached, downloading from url...")
        download_dataset(dataset, local_path)
        return load_dataset(name)


def load_local_dataset(dataset, local_path):
    df = dataset.load_func(local_path)
    if dataset.processing_func:
        df = dataset.processing_func(df)
    return df.iloc[:, :-1].values, df.iloc[:, -1].values


def download_dataset(dataset, local_path):
    r = requests.get(dataset.url)

    if dataset.file_to_unpack is not None:
        with tempfile.TemporaryFile() as fp:
            fp.write(r.content)
            unpack_file_from_zip(fp, dataset.file_to_unpack, local_path)
    else:
        with open(local_path, "wb") as f:
            f.write(r.content)
    _logger.info(f"Downloaded {dataset.name} to {local_path}")


def unpack_file_from_zip(zip_file, to_unpack, path):
    with zipfile.ZipFile(zip_file, "r") as inzipfile:
        with tempfile.TemporaryDirectory() as temp_dir:
            inzipfile.extractall(path=temp_dir)
            data_file = glob.glob(f"{temp_dir}/**/{to_unpack}", recursive=True)[0]
            shutil.copyfile(data_file, path)


# -------------------------------------------------------------------------------------
# Metrics to evaluate
# -------------------------------------------------------------------------------------


def root_mean_squared_error(y_pred, y_test):
    return np.sqrt(mean_squared_error(y_pred, y_test))


def normal_nll(loc, scale, y_test):
    return -norm.logpdf(y_test.flatten(), loc=loc, scale=scale).mean()


@dataclass
class EvalResult:
    total_time: float
    rmse: float
    nll: float


# -------------------------------------------------------------------------------------
# Evaluate decorator, which evaluates the pattern function(split_data) -> predictions
# -------------------------------------------------------------------------------------


@dataclass
class SplitData:
    X_train: np.ndarray
    y_train: np.ndarray
    X_val: np.ndarray
    y_val: np.ndarray
    X_test: np.ndarray
    y_test: np.ndarray


evaluations = []


def evaluate(evaluation_func):
    @wraps(evaluation_func)
    def measured(data):
        t0 = time.perf_counter()
        preds = evaluation_func(data)
        elapsed = time.perf_counter() - t0

        if isinstance(preds, np.ndarray):
            rmse = root_mean_squared_error(preds, data.y_test)
            nll = None
        else:
            rmse = root_mean_squared_error(preds.loc, data.y_test)
            nll = normal_nll(preds.loc, preds.scale, data.y_test)
        return EvalResult(rmse=rmse, nll=nll, total_time=elapsed)

    evaluations.append(measured)
    return measured


# -------------------------------------------------------------------------------------
# Model functions, everything decorated by @evaluate will be evaluated
# -------------------------------------------------------------------------------------


@evaluate
def ngb_regressor(data):
    ngb = NGBRegressor(verbose=False)
    ngb.fit(
        data.X_train,
        data.y_train,
        X_val=data.X_val,
        Y_val=data.y_val,
        early_stopping_rounds=10,
    )
    return ngb.pred_dist(data.X_test, max_iter=ngb.best_val_loss_itr)


@evaluate
def xgb_distribution(data):
    xgbd = XGBDistribution(max_depth=3, natural_gradient=True, n_estimators=500)
    xgbd.fit(
        data.X_train,
        data.y_train,
        eval_set=[(data.X_val, data.y_val)],
        early_stopping_rounds=10,
        verbose=False,
    )
    return xgbd.predict(data.X_test)


@evaluate
def xgb_regressor(data):
    xgb = XGBRegressor(max_depth=3, n_estimators=500)
    xgb.fit(
        data.X_train,
        data.y_train,
        eval_set=[(data.X_val, data.y_val)],
        early_stopping_rounds=10,
        verbose=False,
    )
    return xgb.predict(data.X_test)


# -----------------------------------------------------------------------------
# Functions for running cross-validation becnhmark experiment
# -----------------------------------------------------------------------------


def parse_args():
    argparser = ArgumentParser()
    argparser.add_argument("--data-dir", type=str, default=DATA_DIR)
    argparser.add_argument("--dataset", type=str, default="concrete")
    argparser.add_argument("--random-seed", type=int, default=42)
    argparser.add_argument("--n-folds", type=int, default=10)
    argparser.add_argument("--db-name", type=str, default="results.db")
    return argparser.parse_args()


def run_experiment():
    args = parse_args()
    setup_logging()

    _logger.info(f"Storing datasets and results in {args.data_dir}")
    os.makedirs(args.data_dir, exist_ok=True)

    db = DataBase(data_dir=args.data_dir, db_name=args.db_name)
    _logger.info("Connected to database")

    X, y = load_dataset(args.dataset, data_dir=args.data_dir)
    _logger.info(f"Loaded dataset: `{args.dataset}`, X.shape={X.shape}")

    _logger.info(f"Setting random seed to {args.random_seed}")
    np.random.seed(args.random_seed)

    kf = KFold(n_splits=args.n_folds, shuffle=True, random_state=args.random_seed)
    _logger.info(f"Cross-validation with {args.n_folds} folds...")
    results = _cross_validate_evaluations(kf, X, y, args)

    _logger.info("Inserting results into database...")
    for eval_func, result in zip(evaluations, results):
        df_metrics = aggregate_results(result)
        df_metrics["dataset"] = args.dataset
        df_metrics["model"] = eval_func.__name__
        db.insert_metrics_pdf(df_metrics)

    df_metrics = db.get_metrics_pdf()
    df_summary = summarize_metrics(df_metrics)
    _logger.info(f"\n{df_summary}")


def _cross_validate_evaluations(kfold, X, y, args):
    results = [list() for eval in evaluations]

    for ii, (train_index, test_index) in enumerate(kfold.split(X)):
        _logger.info(f"Fold {ii+1} / {args.n_folds}...")

        X_train, X_test = X[train_index], X[test_index]
        y_train, y_test = y[train_index], y[test_index]
        X_train, X_val, y_train, y_val = train_test_split(
            X_train, y_train, test_size=0.1, random_state=args.random_seed
        )
        split_data = SplitData(
            X_train=X_train,
            y_train=y_train,
            X_val=X_val,
            y_val=y_val,
            X_test=X_test,
            y_test=y_test,
        )
        for eval_func, result in zip(evaluations, results):
            result.append(eval_func(split_data))
    return results


def aggregate_results(results):
    return (
        pd.DataFrame(results)
        .agg(["mean", "std"], axis=0)
        .melt(ignore_index=False, var_name="metric")
        .rename_axis("agg_func")
        .reset_index()
    )


def summarize_metrics(df_metrics):
    return df_metrics.pivot_table(
        values=["value"],
        index=["dataset", "agg_func"],
        columns=["model", "metric"],
    )


# -------------------------------------------------------------------------------------
# Database for storing results of each experiment run
# -------------------------------------------------------------------------------------


class DataBase:
    def __init__(self, data_dir=DATA_DIR, db_name="results.db"):
        _logger.info(f"Using SQLite database at {data_dir}/{db_name}")
        self.metadata = sa.MetaData()
        self.engine = sa.create_engine(f"sqlite:///{data_dir}/{db_name}")
        self.connection = self.engine.connect()
        self._create_tables()

    def _create_tables(self):
        self.metrics = sa.Table(
            "metrics",
            self.metadata,
            sa.Column("dataset", sa.String(), nullable=False),
            sa.Column("model", sa.String(), nullable=False),
            sa.Column("metric", sa.String(), nullable=False),
            sa.Column("agg_func", sa.String(), nullable=False),
            sa.Column("value", sa.Float(), nullable=True),
            sa.Column("created_on", sa.DateTime(), default=datetime.now),
        )
        self.metadata.create_all(self.engine)

    def insert_records(self, records, table="metrics"):
        ins = getattr(self, table).insert()
        self.connection.execute(ins, records)

    def insert_metrics_pdf(self, df):
        self.insert_records(df.to_dict("records"), table="metrics")

    def get_metrics_pdf(self):
        return pd.read_sql(
            sa.select(self.metrics).order_by(self.metrics.c.created_on.desc()),
            self.connection,
        )


def setup_logging(loglevel=logging.INFO):
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(
        level=loglevel, stream=sys.stdout, format=logformat, datefmt="%Y-%m-%d %H:%M:%S"
    )


if __name__ == "__main__":
    run_experiment()



"""LogNormal distribution
"""
import numpy as np
from scipy.stats import lognorm

from xgboost_distribution.distributions.base import BaseDistribution
from xgboost_distribution.distributions.utils import check_is_gt_zero


class LogNormal(BaseDistribution):
    """LogNormal distribution with log scoring.

    Definition:

        f(x) = exp( -[ (log(x) - log(scale)) / (2 s^2) ]^2 / 2 ) / s


    with parameters (scale, s).

    We reparameterize:
        s -> log(s) = a
        scale -> log(scale) = b

    Note that b essentially becomes the 'loc' of the distribution:

        log(x/scale) / s = ( log(x) - log(scale) ) / s

    which can then be taken analogous to the normal distribution's

        (x - loc) / scale

    Hence we can re-use the computations in `distribution.normal`, exchanging:

        y -> log(y)
        scale -> s

    """

    @property
    def params(self):
        return ("scale", "s")

    def check_target(self, y):
        check_is_gt_zero(y)

    def gradient_and_hessian(self, y, params, natural_gradient=True):
        """Gradient and diagonal hessian"""

        log_y = np.log(y)

        loc, log_s = self._split_params(params)  # note loc = log(scale)
        var = np.exp(2 * log_s)

        grad = np.zeros(shape=(len(y), 2))
        grad[:, 0] = (loc - log_y) / var
        grad[:, 1] = 1 - ((loc - log_y) ** 2) / var

        if natural_gradient:
            fisher_matrix = np.zeros(shape=(len(y), 2, 2))
            fisher_matrix[:, 0, 0] = 1 / var
            fisher_matrix[:, 1, 1] = 2

            grad = np.linalg.solve(fisher_matrix, grad)

            hess = np.ones(shape=(len(y), 2))  # we set the hessian constant
        else:
            hess = np.zeros(shape=(len(y), 2))  # diagonal elements only
            hess[:, 0] = 1 / var
            hess[:, 1] = 2 * ((log_y - loc) ** 2) / var

        return grad, hess

    def loss(self, y, params):
        scale, s = self.predict(params)
        return "LogNormalError", -lognorm.logpdf(y, s=s, scale=scale).mean()

    def predict(self, params):
        log_scale, log_s = self._split_params(params)
        scale, s = np.exp(log_scale), np.exp(log_s)

        return self.Predictions(scale=scale, s=s)

    def starting_params(self, y):
        log_y = np.log(y)
        return np.mean(log_y), np.log(np.std(log_y))

    def _split_params(self, params):
        """Return log_scale (loc) and log_s from params"""
        return params[:, 0], params[:, 1]



import pytest

import numpy as np
import pandas as pd

from xgboost_distribution.distributions import Poisson


@pytest.fixture
def poisson():
    return Poisson()


def test_target_validation(poisson):
    valid_target = np.array([0, 1, 4, 5, 10])
    poisson.check_target(valid_target)


@pytest.mark.parametrize(
    "invalid_target",
    [np.array([-0.1, 1.2]), pd.Series([1.1, 0.4, 2.3])],
)
def test_target_validation_raises(poisson, invalid_target):
    with pytest.raises(ValueError):
        poisson.check_target(invalid_target)


@pytest.mark.parametrize(
    "y, params, natural_gradient, expected_grad",
    [
        (
            np.array([1, 1]),
            np.array([np.log(1), np.log(2)]),
            True,
            np.array([[0], [0.5]]),
        ),
        (
            np.array([1, 1]),
            np.array([np.log(1), np.log(2)]),
            False,
            np.array([[0], [1.0]]),
        ),
    ],
)
def test_gradient_calculation(poisson, y, params, natural_gradient, expected_grad):
    grad, hess = poisson.gradient_and_hessian(
        y, params, natural_gradient=natural_gradient
    )
    np.testing.assert_array_equal(grad, expected_grad)


def test_loss(poisson):
    loss_name, loss_value = poisson.loss(
        # fmt: off
        y=np.array([1, ]),
        params=np.array([[np.log(1)], ]),
    )
    assert loss_name == "PoissonError"
    assert loss_value == 1.0



"""Test the base distribution functionality
"""
import pytest

from xgboost_distribution.distributions import (
    Normal,
    format_distribution_name,
    get_distribution,
    get_distribution_doc,
)


def test_get_distribution():
    name = "normal"
    normal_dist = get_distribution(name)
    assert isinstance(normal_dist, Normal)


def test_get_distribution_raises():
    name = "i-do-not-exist"
    with pytest.raises(ValueError):
        get_distribution(name)


def test_get_distribution_doc():
    distribution_doc = get_distribution_doc()
    assert isinstance(distribution_doc, str)


@pytest.mark.parametrize(
    "name, expected_name",
    [("LogNormal", "log-normal"), ("NegativeBinomial", "negative-binomial")],
)
def test_format_distribution_name(name, expected_name):
    formatted_name = format_distribution_name(name)
    assert formatted_name == expected_name



"""Minimal example of XGBDistribution on Boston Housing dataset
"""
from matplotlib import pyplot as plt
from sklearn.datasets import load_boston
from sklearn.model_selection import train_test_split

from xgboost_distribution import XGBDistribution


def plot_residuals(y_true, y_pred, y_err):
    fig, ax = plt.subplots()
    ax.errorbar(
        y_true,
        y_true - y_pred,
        yerr=y_err,
        marker="o",
        linestyle="None",
        c="k",
        markersize=2.5,
        linewidth=0.5,
    )
    ax.axhline(0, c="k", linestyle="--")
    ax.set_xlabel("y_test")
    ax.set_ylabel("y_test - y_pred")
    plt.show()


def main():
    data = load_boston()
    X, y = data.data, data.target
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=0)

    model = XGBDistribution(
        distribution="normal",
        natural_gradient=True,
        max_depth=2,
        n_estimators=500,
    )
    model.fit(
        X_train,
        y_train,
        eval_set=[(X_test, y_test)],
        early_stopping_rounds=10,
        verbose=False,
    )
    preds = model.predict(X_test)

    plot_residuals(y_true=y_test, y_pred=preds.loc, y_err=preds.scale)


if __name__ == "__main__":
    main()
